use std::net::{Ipv4Addr, SocketAddr};
use std::sync::Arc;
use std::time::Duration;

use async_trait::async_trait;
use tokio::io::{AsyncRead, AsyncWrite};
use tokio::net::TcpStream;

use ricq_core::command::wtlogin::LoginResponse;

use crate::client::NetworkStatus;
use crate::ext::common::after_login;
use crate::{Client, RQError, RQResult};

#[async_trait]
pub trait Connector<S: AsyncRead + AsyncWrite + Send + 'static> {
    async fn connect(&self, client: &Arc<Client>) -> std::io::Result<S>;
}

pub struct DefaultConnector;

#[async_trait]
impl Connector<TcpStream> for DefaultConnector {
    async fn connect(&self, _: &Arc<Client>) -> std::io::Result<TcpStream> {
        tokio::select! {
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(42, 81, 172, 81).into(), 80)) => Ok(conn),
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(114, 221, 148, 59).into(), 14000)) => Ok(conn),
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(42, 81, 172, 147).into(), 443)) => Ok(conn),
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(125, 94, 60, 146).into(), 80)) => Ok(conn),
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(114, 221, 144, 215).into(), 80)) => Ok(conn),
            Ok(conn) = TcpStream::connect(SocketAddr::new(Ipv4Addr::new(42, 81, 172, 22).into(), 80)) => Ok(conn),
            else => Err(std::io::Error::new(std::io::ErrorKind::NotConnected, "NotConnected"))
        }
    }
}

/// 自动重连，在掉线后使用，会阻塞到重连结束
pub async fn auto_reconnect<C, S>(
    client: Arc<Client>,
    credential: Credential,
    interval: Duration,
    max: usize,
    connector: C,
) where
    C: Connector<S>,
    S: AsyncRead + AsyncWrite + Send + 'static,
{
    let mut count = 0;
    loop {
        // 如果不是网络原因掉线，不重连（服务端强制下线/被踢下线/用户手动停止）
        if client.get_status() != (NetworkStatus::NetworkOffline as u8) {
            tracing::warn!(
                "client status: {}, auto_reconnect break",
                client.get_status()
            );
            break;
        }
        client.stop(NetworkStatus::NetworkOffline);
        tracing::error!("client will reconnect after {} seconds", interval.as_secs());
        tokio::time::sleep(interval).await;
        let stream = if let Ok(stream) = connector.connect(&client).await {
            count = 0;
            stream
        } else {
            count += 1;
            if count > max {
                tracing::error!("reconnect_count: {}, break!", count);
                break;
            }
            continue;
        };
        let c = client.clone();
        let handle = tokio::spawn(async move { c.start(stream).await });
        tokio::task::yield_now().await; // 等一下，确保连上了
        if let Err(err) = fast_login(&client, &credential).await {
            // token 可能过期了
            tracing::error!("failed to fast_login: {}, break!", err);
            count += 1;
            if count > max {
                tracing::error!("reconnect_count: {}, break!", count);
                break;
            }
            continue;
        }
        tracing::info!("succeed to reconnect");
        after_login(&client).await;
        handle.await.ok();
    }
}

pub struct Password {
    pub uin: i64,
    pub password: String,
}

pub enum Credential {
    Token(ricq_core::Token),
    Password(Password),
}

/// 用于重连
#[async_trait]
pub trait FastLogin {
    async fn fast_login(&self, client: &Arc<Client>) -> RQResult<()>;
}

#[async_trait]
impl FastLogin for ricq_core::Token {
    async fn fast_login(&self, client: &Arc<Client>) -> RQResult<()> {
        match client.token_login(self.clone()).await? {
            LoginResponse::Success(_) => Ok(()),
            _ => Err(RQError::Other("failed to token_login".into())),
        }
    }
}

#[async_trait]
impl FastLogin for Password {
    async fn fast_login(&self, client: &Arc<Client>) -> RQResult<()> {
        let resp = client.password_login(self.uin, &self.password).await?;
        match resp {
            LoginResponse::Success { .. } => return Ok(()),
            LoginResponse::DeviceLockLogin { .. } => {
                return if let LoginResponse::Success { .. } = client.device_lock_login().await? {
                    Ok(())
                } else {
                    Err(RQError::Other("failed to login".into()))
                };
            }
            _ => return Err(RQError::Other("failed to login".into())),
        }
    }
}

/// 如果你非常确定登录过程中不会遇到验证码，可以用 fast_login
pub async fn fast_login(client: &Arc<Client>, credential: &Credential) -> RQResult<()> {
    match credential {
        Credential::Token(token) => token.fast_login(client).await,
        Credential::Password(password) => password.fast_login(client).await,
    }
}
